
// a simple class for String trees.

import java.util.ArrayList;
import java.util.List;
import java.util.Iterator;


class StrTree implements Iterable<String> {
     
    // We do not provide a way to set the String in an existing node; that's a bad idea.
    class treenode {
	private String data_;
	private int count_;	// optional
	private treenode left_, right_;
	public treenode(String d, treenode l, treenode r) {
		data_ = d; count_ = 1; left_ = l; right_ = r;
	}
	public treenode(String d, int c, treenode l, treenode r) {
		data_ = d; count_ = c; left_ = l; right_ = r;
	}
	public String data() {return data_;}
	public int   count() {return count_;}
	public treenode left()  {return left_;}
	public treenode right() {return right_;}
	public void setcount(int count) {count_ = count;}
	public void setleft  (treenode t) {left_ = t;}
	public void setright (treenode t) {right_ = t;}
    }

    private treenode root_ = null;

    // must not be called with val == null
    public void insert(String val) {
	if (root_ == null) {
		root_ = new treenode(val, null, null);
		return;
	} else {
		rinsert(val, root_);
	}
    }

    // recursive insert: does not get called with t==null
    private void rinsert(String val, treenode t) {
	int comp = val.compareTo(t.data()); // = 0 if Strings ==, <0 if val<t.data, etc
	if (comp == 0 ) return;	// This is where you would update the count
	if (comp < 0) {		// val < t.data()
	    if (t.left() == null) {
		    t.setleft(new treenode(val, null, null));
	    } else {
		    rinsert(val, t.left());
	    }
	} else { // val > t.data()
	    if (t.right() == null) {
		    t.setright(new treenode(val, null, null));
	    } else {
		    rinsert(val, t.right());
	    }
	}
    }

    // ============================MyEnumerator2============================

    // traditional recursive traversal, with depth parameter

    public void inTraverse() {inTraverse(root_, 0);}

    private void inTraverse(treenode p, int depth) {
	int indent = 4;
	if (p==null) return;
	inTraverse(p.left(), depth+1);
	for (int i = 0; i< depth*indent; i++) System.out.print(' ');
	System.out.println(p.data());
	inTraverse(p.right(), depth+1);
    }


    /*
    // traverse using an iterator (enumerator)
    // System.Collections.IEnumerable
    public IEnumerable MyEnumerator()  {
	foreach (String s in MyEnumerator(root_)) yield return s;
    }

    private static IEnumerable MyEnumerator(treenode p) {
	if (p==null) yield break;
	foreach (String s in MyEnumerator(p.left())) {yield return s;}
	yield return p.data();
	foreach (String s in MyEnumerator(p.right())) {yield return s;}
    }
    /* */

    // iterative traverse using a stack, simplified algorithm
    public void inTraverseStack() {
  	Stack<treenode> s = new Stack<treenode>();
	treenode n = root_;
	while (true) {
	    while (n!= null) {
		s.push(n);
		n = n.left();
	    }
	    if (s.isEmpty()) break;
	    n = s.pop();
	    System.out.println(n.data());
	    n = n.right();
	} 	
    }

    // iterative traverse using a stack of Objects, general algorithm
    public void inTraverseStackAlt() {
  	Stack<Object> s = new Stack<Object>();
    	s.push(root_);
    	while (!s.isEmpty()) {
            Object x = s.pop();
            if (x instanceof treenode) {
                treenode t = (treenode) x;
                if (t.right() != null) s.push(t.right());
                s.push(t.data());
                if (t.left() != null) s.push(t.left());
            } else { // x is String
                String str = (String) x;
                System.out.println(str);
            }
        }
    }

    // iterative traverse using a queue
    // what does this do?
    public void inTraverseQueue() {
  	Queue<treenode> s = new Queue<treenode>();
	treenode n = root_;
	while (true) {
	    while (n!= null) {
		s.push(n);
		n = n.left();
	    }
	    if (s.isEmpty()) break;
	    n = s.pop();
	    System.out.println(n.data());
	    n = n.right();
	} 
	
    }


    // iterative traverse using a queue of Objects
    public void inTraverseQueueAlt() {
  	Queue<Object> s = new Queue<Object>();
    	s.push(root_);
    	while (!s.isEmpty()) {
            Object x = s.pop();
            if (x instanceof treenode) {
                treenode t = (treenode) x;
                if (t.left() != null) s.push(t.left());
                s.push(t.data());
                if (t.right() != null) s.push(t.right());
            } else { 
                String str = (String) x;
                System.out.println(str);
            }
        }
    }


    /*
    // iterator using a stack
    public IEnumerable MyEnumerator2() {
  	Stack<treenode> s = new Stack<treenode>();
	treenode n = root_;
	while (true) {
	    while (n!= null) {
		s.push(n);
		n = n.left();
	    }
	    if (s.isEmpty()) break;
	    n = s.pop();
	    yield return n.data();
	    n = n.right();
	} 
    }
    /* */

    // next is a way of iteration using NextSetup() and Next()

    private Stack<treenode> stack;
    private treenode current;

    public void nextSetup() {
	stack = new Stack<treenode>();
	current = root_;
	/*
	while (current != null) {
		stack.push(current);
		current = current.left();
	}
	/* */
    }

    public String next() {
	while (current != null) {
		stack.push(current);
		current = current.left();
	}
	if (stack.isEmpty()) return null;
	current = stack.pop();
	String theString = current.data();
	current = current.right();
	return theString;
    }
	
    public Iterator<String> iterator() {
        return new StrTreeIterator();
    }
    
    
    private class StrTreeIterator implements Iterator<String> {
        private Stack<treenode> stack;
        private treenode current;
        private boolean isMore;
        private String nextStr;


        public StrTreeIterator() {
            System.out.println("initializing StrTreeIterator");
            if (root_ == null) {
                isMore = false;
            }else {
                isMore = true;
	            stack = new Stack<treenode>();
	            current = root_;
	            nextStr = findNext();
	        }
        }

        public boolean hasNext() {
            return isMore;
        }

        public String next() {
            String retval = nextStr;
            nextStr = findNext();
            if (nextStr == null) isMore = false;
            return retval;
        }
        
        private String findNext() {
        	while (current != null) {
        		stack.push(current);
        		current = current.left();
        	}
        	if (stack.isEmpty()) {
        	    return null;
        	} else {
            	current = stack.pop();
            	String theString = current.data();
            	current = current.right();
            	return theString;
            }
        }

        public void remove() {
            throw new UnsupportedOperationException();
        }
    }

}


class Stack<T> {

	private ArrayList<T> data_;
	
	public Stack() {
		data_ = new ArrayList<T>();
	}

	public void push(T val) {
		data_.add(val);
	}

	public T pop() {
		T val = data_.get(data_.size()-1);
		data_.remove(data_.size()-1);
		return val;
	}

	public boolean isEmpty() {
		return data_.size() == 0;
	}
}

class Queue<T> {

	private ArrayList<T> data_;
	
	public Queue() {
		data_ = new ArrayList<T>();
	}

	public void push(T val) {
		data_.add(val);
	}

	public T pop() {
		T val = data_.get(0);
		data_.remove(0);	// not efficient
		return val;
	}
	
	public boolean isEmpty() {
		return data_.size() == 0;
	}
}
